/* Dawicontrol generic Linux driver header, 13.09.2023

   Copyright (C) 2022-2023 Dawicontrol Computersysteme GmbH

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License version 2 as published
   by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#define my_version "5.38.01"
#define max_cdblen 16

#define srb_version 2
#define srb_flag_dir_none 0
#define srb_flag_dir_in 1
#define srb_flag_dir_out 2
#define srb_flag_dir_any 3

struct SCSI_req_blk {
    u16     len;
    u16     ver;
    u32     flags;
    u8      reqstatus;
    u8      hoststatus;
    u8      scsistatus;
    u8      message;
    u8      bus;
    u8      target;
    u8      lun;
    u8      cdblen;
    char    cdb[max_cdblen];
    u32     xfersize;
    u32     sg_count;
    void    *sg_list;
    void    *sensebuffer;
    void    *ref;
    void    *extension;
};


/****************************************************************************/
/*      Internal debug functions                                            */
/****************************************************************************/

#ifdef debug_mode
static void walk_sg(struct scatterlist *sg)
{
    int i=0;
    if(sg) {
        printk(KERN_DEBUG "\n");
        printk(KERN_DEBUG " SG:  CPU address         Size     DMA address         Size\n");
    }
    while (sg) {
        printk(KERN_DEBUG "%3d:  0x%px  0x%.5x  0x%px  0x%.5x\n",
                    i, sg_virt(sg), sg->length, (void *)sg_dma_address(sg), sg_dma_len(sg));
        sg = sg_next(sg);
        i++;
    }
    return;
}
#endif


/****************************************************************************/
/*      External functions                                                  */
/****************************************************************************/

extern int BIN_init_driver(int timer_size, void *sgtbl_size, gfp_t f1, gfp_t f2);
extern int BIN_init_hba(struct pci_dev *pdev, struct Scsi_Host *host, void *hba);
extern int BIN_remove_hba(struct pci_dev *pdev, void *hba);
extern int BIN_suspend_hba(struct pci_dev *pdev, void *hba);
extern int BIN_resume_hba(struct pci_dev *pdev, void *hba);
extern void BIN_shutdown_hba(struct pci_dev *pdev, void *hba, int flags);
extern irqreturn_t BIN_Irq_Handler(int irq, void *hba);
extern int BIN_exec_req(struct SCSI_req_blk *srb, void *hba);
extern int BIN_bus_reset(void *hba, int bus);
extern int BIN_hba_reset(void *hba);
extern int BIN_get_hba (void *buffer, void *hn);
extern int BIN_get_device(void *buffer, void *hn, void *dn);
extern int BIN_get_raid(void *buffer, void *hn, void *rn);


/****************************************************************************/
/*      Commandline parameters                                              */
/****************************************************************************/

//***** disable NCQ support, regardless of BIOS-Setting

bool MOD_ncq_para = 1;
module_param_named(ncq, MOD_ncq_para, bool, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(ncq, " disable NCQ support: nqc=0 (default=BIOS setting)");


//***** disable TRIM support

bool MOD_trim_para = 1;
module_param_named(trim, MOD_trim_para, bool, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(trim, " disable TRIM/UNMAP support: trim=0 (default=enable)");


//***** disable sleepmode, regardless of BIOS-Settings

bool MOD_sleepmode_para = 1;
module_param_named(sleepmode, MOD_sleepmode_para, bool, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(sleepmode, " disable sleepmode of drives: sleepmode=0 (default=BIOS setting)");


//***** force writecache of drives, regardless of BIOS- and System-Settings

bool MOD_forcewrcache_para = 0;
module_param_named(forcewrcache, MOD_forcewrcache_para, bool, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(forcewrcache, " force writecache of drives: forcewrcache=1 (default=system setting)");


//***** delay start of pending rebuild-job (if any)

int MOD_rebuilddelay_para = 2;
module_param_named(rebuilddelay, MOD_rebuilddelay_para, int, S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP);
MODULE_PARM_DESC(rebuilddelay, " delay rebuild-job: rebuilddelay=m (m=0~60, default=2 minutes)");


/****************************************************************************/
/*      Callback functions                                                  */
/****************************************************************************/

//***** Print message

void MOD_print(char *message)
{
    printk("%s", message);
    return;
}

//***** Allocate / release non-DMA memory

void *MOD_allocate_memory(unsigned long size)
{
    return vmalloc(size);
}

void MOD_release_memory(void *vadr)
{
    vfree(vadr);
    return;
}

//***** Install / uninstall interrupt handler

int MOD_install_irq(struct pci_dev *pdev, void *hba, unsigned int *devirq)
{
    *devirq = pdev->irq;  //return the irq#, maybe different from bar value
    return request_irq(pdev->irq, BIN_Irq_Handler, IRQF_SHARED, my_name, hba);
}

void MOD_uninstall_irq(unsigned int irq, void *hba)
{
    free_irq(irq, hba);
    return;
}

//***** Return scsi_cmnd

void MOD_return_req(struct SCSI_req_blk *srb)
{
    struct scsi_cmnd *cmdblk;

    // get scsi_cmnd and update status
    cmdblk = srb->ref;
    cmdblk->result = (srb->reqstatus << 24) + (srb->hoststatus << 16)
                   + (srb->message << 8) + (srb->scsistatus);

    scsi_set_resid(cmdblk, scsi_bufflen(cmdblk) - srb->xfersize);

    if (srb->sg_count) {
        scsi_dma_unmap(cmdblk);
    }
    // log return parameters
    #ifdef debug_mode
        // printk(KERN_DEBUG "\n");
        // printk(KERN_DEBUG my_name ": Srb = 0x%px return on Scsi_Cmnd = 0x%px \n", srb, cmdblk);
        if (cmdblk->result) {
            printk(KERN_ERR my_name ": Return: ReqStatus = 0x%.2x,  HostStatus = 0x%.2x,  ScsiStatus = 0x%.2x\n",
                            srb->reqstatus, srb->hoststatus, srb->scsistatus);
        }
    #endif

    // return command to upper layer
    #if LINUX_VERSION_CODE >= KERNEL_VERSION(5,16,0)
        scsi_done(cmdblk);
    #else
        cmdblk->scsi_done(cmdblk);
    #endif

    return;
 }

//***** Scan host

void MOD_scan_host(struct Scsi_Host *host)
{
    scsi_scan_host(host);
    return;
}

/****************************************************************************/
/*      Delay & Timer functions                                             */
/****************************************************************************/

//***** Wait uSeconds

void MOD_wait_usecs(ulong value)
{
    udelay(value);
    return;
}

//***** Install / uninstall / start timer

void MOD_install_timer (struct timer_list *this_timer, void *callback)
{
    timer_setup(this_timer, callback, 0);
    return;
}

void MOD_uninstall_timer(struct timer_list *this_timer)
{
    del_timer_sync(this_timer);
    return;
}

int MOD_start_timer(struct timer_list *this_timer, ulong period)
{
    return mod_timer(this_timer, jiffies + msecs_to_jiffies(period));
}

/****************************************************************************/
/*      DMA related functions                                               */
/****************************************************************************/

//***** Setup DMA 32/64 bit mask

int MOD_setup_dma(struct pci_dev *pdev)
{
    int error;

    if (!dma_set_mask(&pdev->dev, DMA_BIT_MASK(64))) {
        error = dma_set_coherent_mask(&pdev->dev, DMA_BIT_MASK(64));
        if (error) {
            printk(KERN_ERR my_name ": setup 64-bit coherent DMA failed.\n");
            error = dma_set_coherent_mask(&pdev->dev, DMA_BIT_MASK(32));
            if (error) {
                printk(KERN_ERR my_name ": setup 32-bit coherent DMA failed.\n");
                return error;
            }
        }
    } else {
        printk(KERN_ERR my_name ": setup 64-bit DMA failed.\n");
        error = dma_set_mask(&pdev->dev, DMA_BIT_MASK(32));
        if (error) {
            printk(KERN_ERR my_name ": setup 32-bit DMA failed.\n");
            return error;
        }
        error = dma_set_coherent_mask(&pdev->dev, DMA_BIT_MASK(32));
        if (error) {
            printk(KERN_ERR my_name ": setup 32-bit coherent DMA failed.\n");
            return error;
        }
    }
    return error;
}

//***** Allocate / release common buffer

void *MOD_allocate_common(struct pci_dev *pdev, size_t size, dma_addr_t *padr, gfp_t flags)
{
    return dma_alloc_coherent(&pdev->dev, size, padr, flags);
}

void MOD_release_common(struct pci_dev *pdev, size_t size, void *vadr, dma_addr_t padr)
{
    dma_free_coherent(&pdev->dev, size, vadr, padr);
    return;
}

//***** get SGe content

struct scatterlist *MOD_get_sge(struct scatterlist *sge, dma_addr_t *padr,
                unsigned long *vadr, unsigned int *vlen, unsigned int *plen)
{
    *padr = sg_dma_address(sge);
    *vadr = (unsigned long) sg_virt(sge);
    *vlen = sge->length;
    *plen = sg_dma_len(sge);
    return sg_next(sge);
}

/****************************************************************************/
/*      Access memory-mapped I/O-space                                      */
/****************************************************************************/

//***** Request / release MMIO-region

int MOD_request_mmio (ulong padr, ulong size)
{
    int error = 0;

    if (!request_mem_region(padr, size, my_name)) {
        error = -EBUSY;
    }
    return error;
}

void MOD_release_mmio (ulong padr, ulong size)
{
    return release_mem_region(padr, size);
}

//***** Map MMIO-space to virtual address

void *MOD_map_mmio(resource_size_t padr, ulong size)
{
    return ioremap(padr, size);
}

//***** Read / write memory-mapped IO-space

u32 MOD_read_mapped(void __iomem *vadr)
{
    return readl(vadr);
}

void MOD_write_mapped(void __iomem *vadr, u32 value)
{
    writel (value, vadr);
    return;
}

/****************************************************************************/
/*      Access PCI-space                                                    */
/****************************************************************************/

//***** Enable / disable PCI device

int MOD_enable_pci(struct pci_dev *pdev, unsigned int *dadr)
{
    // get pci device address
    // 15:8 = bus, 7:3 =device, 2:0 = function
    *dadr = (pdev->bus->number << 8) + (pdev->devfn);

    // 31:16 = domain
    #ifdef CONFIG_PCI_DOMAINS_GENERIC
        *dadr = *dadr + (pdev->bus->domain_nr << 16);
    #endif

    // enable pci device
    return pci_enable_device(pdev);
}

void MOD_disable_pci(struct pci_dev *pdev)
{
    pci_disable_device(pdev);
    return;
}

//***** Read / write PCI config register

u32 MOD_read_config(struct pci_dev *pdev, int bar)
{
    u32 value;

    pci_read_config_dword(pdev, bar, &value);
    return value;
}

void MOD_write_config(struct pci_dev *pdev, int bar, u32 value)
{
    pci_write_config_dword(pdev, bar, value);
    return;
}
